@@ -7,6 +7,8 @@ Author: Florent Xicluna
 """
 from __future__ import with_statement
 
+import _ast
+import collections
 import functools
 import optparse
 import os
@@ -24,48 +26,12 @@ try:                    # Python 3
     int_types = int
 except ImportError:     # Python 2
     import ConfigParser as configparser
-    from itertools import ifilter as filter
     from threading import currentThread as current_thread
     from xmlrpclib import Fault, ServerProxy
     int_types = int, long
 
-try:
-    from ast import literal_eval
-except ImportError:     # Python 2.5
-    import _ast
 
-    # Port of Python 2.6's ast.literal_eval for use under Python 2.5
-    def _convert(node, _consts={'None': None, 'True': True, 'False': False}):
-        if isinstance(node, _ast.Str):
-            return node.s
-        if isinstance(node, _ast.Num):
-            return node.n
-        if isinstance(node, _ast.Tuple):
-            return tuple(map(_convert, node.elts))
-        if isinstance(node, _ast.List):
-            return list(map(_convert, node.elts))
-        if isinstance(node, _ast.Dict):
-            return dict([(_convert(k), _convert(v))
-                         for (k, v) in zip(node.keys, node.values)])
-        if isinstance(node, _ast.Name) and node.id in _consts:
-            return _consts[node.id]
-        raise ValueError('malformed or disallowed expression')
-
-    def literal_eval(node_or_string):
-        if isinstance(node_or_string, basestring):
-            node_or_string = compile(node_or_string,
-                                     '<unknown>', 'eval', _ast.PyCF_ONLY_AST)
-        if isinstance(node_or_string, _ast.Expression):
-            node_or_string = node_or_string.body
-        return _convert(node_or_string)
-
-    def next(iterator, default):
-        for item in iterator:
-            return item
-        return default
-
-
-__version__ = '1.4.5'
+__version__ = '1.5'
 __all__ = ['Client', 'Model', 'Record', 'RecordList', 'Service',
            'format_exception', 'read_config', 'start_openerp_services']
 
@@ -77,28 +43,20 @@ DEFAULT_USER = 'admin'
 MAXCOL = [79, 179, 9999]    # Line length in verbose mode
 
 USAGE = """\
-Usage (main commands):
-    search(obj, domain)
-    search(obj, domain, offset=0, limit=None, order=None)
-                                    # Return a list of IDs
-    count(obj, domain)              # Count the matching objects
-
-    read(obj, ids, fields=None)
-    read(obj, domain, fields=None)
-    read(obj, domain, fields=None, offset=0, limit=None, order=None)
-                                    # Return values for the fields
-
+Usage (some commands):
     models(name)                    # List models matching pattern
     model(name)                     # Return a Model instance
-    keys(obj)                       # List field names of the model
-    fields(obj, names=None)         # Return details for the fields
-    field(obj, name)                # Return details for the field
-    access(obj, mode='read')        # Check access on the model
+    model(name).keys()              # List field names of the model
+    model(name).fields(names=None)  # Return details for the fields
+    model(name).field(name)         # Return details for the field
+    model(name).browse(domain)
+    model(name).browse(domain, offset=0, limit=None, order=None)
+                                    # Return a RecordList
 
-    do(obj, method, *params)        # Generic 'object.execute'
-    exec_workflow(obj, signal, id)  # Trigger workflow signal
+    rec = model(name).get(domain)   # Get the Record matching domain
+    rec.some_field                  # Return the value of this field
+    rec.read(fields=None)           # Return values for the fields
 
-    client                          # Client object, connected
     client.login(user)              # Login with another user
     client.connect(env)             # Connect to another env.
     client.modules(name)            # List modules matching pattern
@@ -112,7 +70,7 @@ DOMAIN_OPERATORS = frozenset('!|&')
 #   =, !=, >, >=, <, <=, like, ilike, in, not like, not ilike, not in, child_of
 #   =like, =ilike (6.0), =? (6.0)
 _term_re = re.compile(
-    '([\w._]+)\s*'  '(=(?:like|ilike|\?)|[<>!]=|[<>=]'
+    '([\w._]+)\s*'  '(=(?:like|ilike|\?)|[<>]=?|!?=(?!=)'
     '|(?<= )(?:like|ilike|in|not like|not ilike|not in|child_of))'  '\s*(.*)')
 _fields_re = re.compile(r'(?:[^%]|^)%\(([^)]+)\)')
 
@@ -140,6 +98,37 @@ _cause_message = ("\nThe above exception was the direct cause "
                   "of the following exception:\n\n")
 
 
+# Simplified ast.literal_eval which does not parse operators
+def _convert(node, _consts={'None': None, 'True': True, 'False': False}):
+    if isinstance(node, _ast.Str):
+        return node.s
+    if isinstance(node, _ast.Num):
+        return node.n
+    if isinstance(node, _ast.Tuple):
+        return tuple(map(_convert, node.elts))
+    if isinstance(node, _ast.List):
+        return list(map(_convert, node.elts))
+    if isinstance(node, _ast.Dict):
+        return dict([(_convert(k), _convert(v))
+                     for (k, v) in zip(node.keys, node.values)])
+    if isinstance(node, _ast.Name) and node.id in _consts:
+        return _consts[node.id]
+    raise ValueError('malformed or disallowed expression')
+
+
+def literal_eval(expression):
+    node = compile(expression, '<unknown>', 'eval', _ast.PyCF_ONLY_AST)
+    return _convert(node.body)
+
+
+def is_list_of_dict(iterator):
+    """Return True if the first non-false item is a dict."""
+    for item in iterator:
+        if item:
+            return isinstance(item, dict)
+    return False
+
+
 def mixedcase(s, _cache={}):
     """Convert to MixedCase.
 
@@ -179,7 +168,7 @@ def format_exception(exc_type, exc, tb, limit=None, chain=True,
     if ((issubclass(exc_type, Fault) and
          isinstance(exc.faultCode, basestring))):
         # Format readable 'Fault' errors
-        etype, _, msg = exc.faultCode.partition('--')
+        (etype, __, msg) = exc.faultCode.partition('--')
         server_tb = None
         if etype.strip() != 'warning':
             msg = exc.faultCode
@@ -278,11 +267,11 @@ def searchargs(params, kwargs=None, context=None):
         warnings.warn('Domain should be a list: %s' % domain)
     elif not isinstance(domain, list):
         return params
-    for idx, term in enumerate(domain):
+    for (idx, term) in enumerate(domain):
         if isinstance(term, basestring) and term not in DOMAIN_OPERATORS:
             m = _term_re.match(term.strip())
             if not m:
-                raise ValueError("Cannot parse term %r" % term)
+                raise ValueError('Cannot parse term %r' % term)
             (field, operator, value) = m.groups()
             try:
                 value = literal_eval(value)
@@ -526,8 +515,6 @@ class Client(object):
     def _set_interactive(cls):
         # Don't call multiple times
         del Client._set_interactive
-        global_names = ['wizard', 'exec_workflow', 'read', 'search', 'count',
-                        'model', 'models', 'keys', 'fields', 'field', 'access']
 
         def connect(self, env=None):
             """Connect to another environment and replace the globals()."""
@@ -543,13 +530,12 @@ class Client(object):
             sys.ps2 = '%s ... ' % (env,)
             # Logged in?
             if client.user:
+                g['model'] = client.model
+                g['models'] = client.models
                 g['do'] = client.execute
-                for name in global_names:
-                    g[name] = getattr(client, name)
                 print('Logged in as %r' % (client.user,))
             else:
-                g['do'] = None
-                g.update(dict.fromkeys(global_names))
+                g['model'] = g['models'] = g['do'] = None
 
         def login(self, user, password=None, database=None):
             """Switch `user` and (optionally) `database`."""
@@ -593,7 +579,7 @@ class Client(object):
         assert isinstance(obj, basestring)
         assert isinstance(method, basestring) and method != 'browse'
         context = kwargs.pop('context', None)
-        ordered = None
+        ordered = single_id = False
         if method in ('read', 'name_get'):
             assert params
             if issearchdomain(params[0]):
@@ -605,13 +591,14 @@ class Client(object):
                 ordered = kwargs.pop('order', False) and params[0]
                 ids = set(params[0])
                 ids.discard(False)
-                if not ids:
+                if not ids and ordered:
                     return [False] * len(ordered)
                 ids = sorted(ids)
             else:
-                ids = params[0]
+                single_id = True
+                ids = [params[0]] if params[0] else False
             if not ids:
-                return []
+                return ids
             if len(params) > 1:
                 params = (ids,) + params[1:]
             elif method == 'read':
@@ -641,7 +628,7 @@ class Client(object):
             if not isinstance(ordered, list):
                 ordered = ids
             res = [resdic.get(id_, False) for id_ in ordered]
-        return res
+        return res[0] if single_id else res
 
     def exec_workflow(self, obj, signal, obj_id):
         """Wrapper around ``object.exec_workflow`` RPC method.
@@ -788,14 +775,11 @@ class Client(object):
             return res[fields[0]]
         return res
 
-    def _model(self, name):
+    def _models_get(self, name):
         try:
             return self._models[name]
         except KeyError:
-            # m = Model(self, name)
-            m = object.__new__(Model)
-        m._init(self, name)
-        self._models[name] = m
+            self._models[name] = m = Model._new(self, name)
         return m
 
     def models(self, name=''):
@@ -814,7 +798,7 @@ class Client(object):
         domain = [('model', 'like', name)]
         models = self.execute('ir.model', 'read', domain, ('model',))
         names = [m['model'] for m in models]
-        return dict([(mixedcase(name), self._model(name)) for name in names])
+        return dict([(mixedcase(mod), self._models_get(mod)) for mod in names])
 
     def model(self, name, check=True):
         """Return a :class:`Model` instance.
@@ -823,7 +807,7 @@ class Client(object):
         argument `check` is :const:`False`, no validity check is done.
         """
         try:
-            return self._models[name] if check else self._model(name)
+            return self._models[name] if check else self._models_get(name)
         except KeyError:
             models = self.models(name)
         if name in self._models:
@@ -851,12 +835,9 @@ class Client(object):
             domain.append(('state', op, ['uninstalled', 'uninstallable']))
         mods = self.read('ir.module.module', domain, 'name state')
         if mods:
-            res = {}
+            res = collections.defaultdict(list)
             for mod in mods:
-                if mod['state'] in res:
-                    res[mod['state']].append(mod['name'])
-                else:
-                    res[mod['state']] = [mod['name']]
+                res[mod['state']].append(mod['name'])
             return res
 
     def keys(self, obj):
@@ -903,13 +884,15 @@ class Model(object):
     def __new__(cls, client, name):
         return client.model(name)
 
-    def _init(self, client, name):
-        self.client = client
-        self._name = name
-        self._execute = functools.partial(client.execute, name)
-        self.search = functools.partial(client.search, name)
-        self.count = functools.partial(client.count, name)
-        self.read = functools.partial(client.read, name)
+    @classmethod
+    def _new(cls, client, name):
+        m = object.__new__(cls)
+        (m.client, m._name) = (client, name)
+        m._execute = functools.partial(client.execute, name)
+        m.search = functools.partial(client.search, name)
+        m.count = functools.partial(client.count, name)
+        m.read = functools.partial(client.read, name)
+        return m
 
     def __repr__(self):
         return "<Model '%s'>" % (self._name,)
@@ -985,14 +968,14 @@ class Model(object):
         if isinstance(domain, int_types):   # a single id
             return Record(self, domain, context=context)
         if isinstance(domain, basestring):  # lookup the xml_id
-            module, name = domain.split('.')
-            data = self.client.model('ir.model.data').read(
+            (module, name) = domain.split('.')
+            data = self._imd_read(
                 [('module', '=', module), ('name', '=', name)], 'model res_id')
             assert not data or data[0]['model'] == self._name
             ids = [res['res_id'] for res in data]
         else:                               # a search domain
             assert issearchdomain(domain)
-            params = searchargs((domain,), context=context)
+            params = searchargs((domain,), {}, context)
             ids = self._execute('search', *params)
         if len(ids) > 1:
             raise ValueError('domain matches too many records (%d)' % len(ids))
@@ -1016,8 +999,8 @@ class Model(object):
         the value is wrapped in a Record or a RecordList.
         Return a dictionary with the same keys as the `values` argument.
         """
-        for key, value in values.items():
-            if key == 'id':
+        for (key, value) in values.items():
+            if key == 'id' or hasattr(value, 'id'):
                 continue
             field = self._fields[key]
             field_type = field['type']
@@ -1029,7 +1012,7 @@ class Model(object):
                 rel_model = self.client.model(field['relation'], False)
                 values[key] = RecordList(rel_model, value, context=context)
             elif value and field_type == 'reference':
-                res_model, res_id = value.split(',')
+                (res_model, res_id) = value.split(',')
                 rel_model = self.client.model(res_model, False)
                 values[key] = Record(rel_model, int(res_id))
         return values
@@ -1037,9 +1020,9 @@ class Model(object):
     def _unbrowse_values(self, values):
         """Unwrap the id of Record and RecordList."""
         new_values = values.copy()
-        for key, value in values.items():
+        for (key, value) in values.items():
             field_type = self._fields[key]['type']
-            if isinstance(value, (Record, RecordList)):
+            if hasattr(value, 'id'):
                 if field_type == 'reference':
                     new_values[key] = '%s,%s' % (value._model_name, value.id)
                 else:
@@ -1051,10 +1034,29 @@ class Model(object):
                     new_values[key] = [(6, 0, value)]
         return new_values
 
+    def _get_external_ids(self, ids=None):
+        """Retrieve the External IDs of the records.
+
+        Return a dictionary with keys being the fully qualified
+        External IDs, and values the ``Record`` entries.
+        """
+        search_domain = [('model', '=', self._name)]
+        if ids is not None:
+            search_domain.append(('res_id', 'in', ids))
+        existing = self._imd_read(search_domain, ['module', 'name', 'res_id'])
+        res = {}
+        for rec in existing:
+            res['%(module)s.%(name)s' % rec] = self.get(rec['res_id'])
+        return res
+
     def __getattr__(self, attr):
         if attr in ('_keys', '_fields'):
             self.__dict__[attr] = rv = getattr(self, '_get' + attr)()
             return rv
+        if attr.startswith('_imd_'):
+            imd = self.client.model('ir.model.data')
+            self.__dict__[attr] = imd_method = getattr(imd, attr[5:])
+            return imd_method
         if attr.startswith('_'):
             raise AttributeError("'Model' object has no attribute %r" % attr)
 
@@ -1079,18 +1081,17 @@ class RecordList(object):
     """
 
     def __init__(self, res_model, ids, context=None):
-        _ids = []
-        for id_ in ids:
+        idnames = list(ids)
+        for (index, id_) in enumerate(ids):
             if isinstance(id_, (list, tuple)):
-                _ids.append(id_[0])
-            else:
-                _ids.append(id_)
+                ids[index] = id_ = id_[0]
+            assert isinstance(id_, int_types), repr(id_)
         # Bypass the __setattr__ method
         self.__dict__.update({
-            'id': _ids,
+            'id': ids,
             '_model_name': res_model._name,
             '_model': res_model,
-            '_idnames': ids,
+            '_idnames': idnames,
             '_context': context,
             '_execute': res_model._execute,
         })
@@ -1104,7 +1105,8 @@ class RecordList(object):
 
     def __dir__(self):
         return ['__getitem__', 'read', 'write', 'unlink', '_context',
-                '_idnames', '_model', '_model_name'] + self._model._keys
+                '_idnames', '_model', '_model_name',
+                '_external_id'] + self._model._keys
 
     def __len__(self):
         return len(self.id)
@@ -1118,7 +1120,7 @@ class RecordList(object):
         if self.id:
             values = client.read(self._model_name, self.id,
                                  fields, order=True, context=context)
-            if isinstance(next(filter(None, values), None), dict):
+            if is_list_of_dict(values):
                 browse_values = self._model._browse_values
                 return [v and browse_values(v) for v in values]
         else:
@@ -1137,7 +1139,7 @@ class RecordList(object):
                     records = []
                     for value in values:
                         if value:
-                            res_model, res_id = value.split(',')
+                            (res_model, res_id) = value.split(',')
                             rel_model = client.model(res_model, False)
                             value = Record(rel_model, int(res_id))
                         records.append(value)
@@ -1163,6 +1165,18 @@ class RecordList(object):
         rv = self._execute('unlink', self.id, context=context)
         return rv
 
+    @property
+    def _external_id(self):
+        """Retrieve the External IDs of the :class:`RecordList`.
+
+        Return the fully qualified External IDs with default value
+        False if there's none.  If multiple IDs exist for a record,
+        only one of them is returned (randomly).
+        """
+        xml_ids = dict([(r.id, xml_id) for (xml_id, r) in
+                        self._model._get_external_ids(self.id).items()])
+        return [xml_ids.get(res_id, False) for res_id in self.id]
+
     def __getitem__(self, key):
         idname = self._idnames[key]
         if idname is False:
@@ -1212,6 +1226,7 @@ class Record(object):
         if isinstance(res_id, (list, tuple)):
             (res_id, res_name) = res_id
             self.__dict__['_name'] = res_name
+        assert isinstance(res_id, int_types), repr(res_id)
         # Bypass the __setattr__ method
         self.__dict__.update({
             'id': res_id,
@@ -1231,9 +1246,10 @@ class Record(object):
     def _get_name(self):
         try:
             (id_name,) = self._execute('name_get', [self.id])
-            name = '[%d] %s' % (self.id, id_name[1])
+            name = '%s' % (id_name[1],)
         except Exception:
-            name = '[%d] -' % (self.id,)
+            name = '%s,%d' % (self._model_name, self.id)
+        self.__dict__['_name'] = name
         return name
 
     @property
@@ -1318,9 +1334,31 @@ class Record(object):
         self.refresh()
         return rv
 
+    @property
+    def _external_id(self):
+        """Retrieve the External ID of the :class:`Record`.
+
+        Return the fully qualified External ID of the :class:`Record`,
+        with default value False if there's none.  If multiple IDs
+        exist, only one of them is returned (randomly).
+        """
+        xml_ids = self._model._get_external_ids([self.id])
+        return list(xml_ids)[0] if xml_ids else False
+
+    def _set_external_id(self, xml_id):
+        """Set the External ID of this record."""
+        (mod, name) = xml_id.split('.')
+        obj = self._model_name
+        domain = ['|', '&', ('model', '=', obj), ('res_id', '=', self.id),
+                  '&', ('module', '=', mod), ('name', '=', name)]
+        if self._model._imd_search(domain):
+            raise ValueError('ID %r collides with another entry' % xml_id)
+        vals = {'model': obj, 'res_id': self.id, 'module': mod, 'name': name}
+        self._model._imd_create(vals)
+
     def __dir__(self):
         return ['read', 'write', 'copy', 'unlink', '_send', 'refresh',
-                '_context', '_model', '_model_name', '_name',
+                '_context', '_model', '_model_name', '_name', '_external_id',
                 '_keys', '_fields'] + self._model._keys
 
     def __getattr__(self, attr):
@@ -1328,8 +1366,7 @@ class Record(object):
         if attr in self._model._keys:
             return self.read(attr, context=context)
         if attr == '_name':
-            self.__dict__['_name'] = name = self._get_name()
-            return name
+            return self._get_name()
         if attr.startswith('_'):
             raise AttributeError("'Record' object has no attribute %r" % attr)
 
@@ -1338,6 +1375,7 @@ class Record(object):
             if context:
                 kwargs.setdefault('context', context)
             res = self._execute(attr, [self.id], *params, **kwargs)
+            self.refresh()
             if isinstance(res, list) and len(res) == 1:
                 return res[0]
             return res
@@ -1347,12 +1385,18 @@ class Record(object):
         return mobj
 
     def __setattr__(self, attr, value):
+        if attr == '_external_id':
+            return self._set_external_id(value)
         if attr not in self._model._keys:
             raise AttributeError("'Record' object has no attribute %r" % attr)
         if attr == 'id':
             raise AttributeError("'Record' object attribute 'id' is read-only")
         self.write({attr: value})
 
+    def __eq__(self, other):
+        return (isinstance(other, Record) and
+                self.id == other.id and self._model_name == other._model_name)
+
 
 def _interact(use_pprint=True, usage=USAGE):
     import code
@@ -1405,7 +1449,7 @@ def _interact(use_pprint=True, usage=USAGE):
             except:
                 # Print readable 'Fault' errors
                 # Work around http://bugs.python.org/issue12643
-                exc_type, exc, tb = sys.exc_info()
+                (exc_type, exc, tb) = sys.exc_info()
                 msg = ''.join(format_exception(exc_type, exc, tb, chain=False))
                 print(msg.strip())
 
